from adoconstants import *
from string import join, split, find, strip, translate, maketrans
import sys, types
from Shared.DC.ZRDB.THUNK import THUNKED_TM
from win32com.client import Dispatch
from time import time
import threading
import pythoncom
rc=sys.getrefcount

error=DatabaseError='ADO Database Error'

defs={
  adSmallInt: u'i',
  adInteger: u'i',
  adSingle: u'n',
  adDouble: u'n',
  adCurrency: u'n',
  adDate: u'd',
  adBSTR: u't',
  adBoolean: u'i',
  adDecimal: u'n',
  adTinyInt: u'i',
  adUnsignedTinyInt: u'i',
  adUnsignedSmallInt: u'i',
  adUnsignedInt: u'i',
  adBigInt: u'i',
  adUnsignedBigInt: u'i',
  adBinary: u't',
  adChar: u't',
  adWChar: u't',
  adNumeric: u'n',
  adDBDate: u'd',
  adDBTime: u'd',
  adDBTimeStamp: u'd',
  adVarNumeric: u'n',
  adVarChar: u't',
  adLongVarChar: u't',
  adVarWChar: u't',
  adLongVarWChar: u't',
  adVarBinary: u't',
  adLongVarBinary: u't',
  }

binary_types=adBinary, adLongVarBinary, adVarBinary

class ConnectionWrapper:
    def __init__(self, pool, con_str, con):
        self._con = con
        self._con_str = con_str
        self._pool = pool

    def __del__(self):
        self.close()
    
    def __getattr__(self, attr):
        if not self._con:
            raise ValueError, "Database connection used after being closed."
        return getattr(self._con, attr)
    
    def close(self):
        pool = self._pool
        if pool:
            self._pool = None
            pool.release(self)

class ConnectionPool:
    def __init__(self, timeout = 60):
        self.connection_map = {}
        self.timeout = timeout
        self.sweeper = threading.Thread(None, self._sweeper)
        self.sentry = threading.Condition()
        self.exiting = 0
        self.sweeper.start()
        
    def __del__(self):
        self.sentry.acquire()
        self.exiting = 1
        self.sentry.notifyAll()
        self.sentry.release()
        
    
    def acquire(self, connection_string):
        self.sentry.acquire()
        try:
            if not self.connection_map.has_key(connection_string):
                self.connection_map[connection_string] = []
            clist = self.connection_map[connection_string]
            if len(clist) == 0:
	      # Create Connection and configure it for client side cursors (faster)
                con = Dispatch('ADODB.Connection')
                con.ConnectionString = connection_string
		con.CursorLocation = adUseClient
                con.Open()
                clist.append((con, time()))
            return ConnectionWrapper(self, connection_string, clist.pop()[0])
        finally:
            self.sentry.release()
    
    def release(self, connection):
        self.sentry.acquire()
        try:
            clist = self.connection_map[connection._con_str]
            clist.append((connection._con, time()))
            connection._pool = None
            connection._con = None
        finally:
            self.sentry.release()
    
    def flush(self, connection_string):
        self.sentry.acquire()
        try:
            clist = self.connection_map[connection_string]
            for c in clist:
                c[0].Close()
            del self.connection_map[connection_string]
        finally:
            self.sentry.release()
    
    def flush_all(self):
        self.sentry.acquire()
        try:
            for cs in self.connection_map.keys():
                clist = self.connection_map[cs]
                for con in clist: con[0].Close()
                del self.connection_map[cs]
        finally:
            self.sentry.release()
            
    def flush_old(self):
        self.sentry.acquire()
        try:
            t = time()
            young = lambda con,t=t,self=self: con[1] - t < self.timeout
            for cs in self.connection_map.keys():
                clist = self.connection_map[cs]
                self.connection_map[cs] = filter(young, clist)
                for con in clist:
                    if not young(con):
                        con[0].Close()
        finally:
            self.sentry.release()

    def _sweeper(self):
        while not self.exiting:
            self.sentry.acquire()
            self.sentry.wait(10)
            self.flush_old()
            self.sentry.release()

ConnectionPool = lambda singleton = ConnectionPool(): singleton


class DB(THUNKED_TM):
  """This class connects actually to the database and performs the query execution"""
  
  def __init__(self, connection_string=''):
    """Initializes the database connection"""
    self.connection_string=connection_string
    self.isConnected = 0
    pythoncom.CoInitialize()
  
  def __del__(self):
    self.disconnect()
    pythoncom.CoUninitialize()

  def connect(self):
    """Connects the database connection to the database"""
    if not self.isConnected:
      self.connection = ConnectionPool().acquire(self.connection_string)
      self.isConnected = 1
      return 0 # was not connected
    return 1 # was connected

  def disconnect(self):
    """Disconnects the database connection from the database"""
    if self.isConnected:
      self.connection.close()
      del self.connection
      self.isConnected = 0
    
  def table_info(self):
    """Returns info about the tables contained in a database"""
    wasConnected = self.connect()
    oCat = Dispatch('ADOX.Catalog')
    oCat.ActiveConnection = self.connection
    oTab = oCat.Tables
    r=[]
    for table in oTab:
      if table.Type == 'TABLE':
	name = table.Name
	try:
	  owner = '(owned by %s)'%oCat.GetObjectOwner(name, adCmdTable)
	except:
	  owner = ''
	r.append((name, owner))
    # release ActiveX Catalog
    oCat = None

    return r

  def query(self, src=None, max_rows=99999999, rdb=1):
    """Executes a query and returns the result"""
    self._register()
    wasConnected = self.connect()

    try:
      if src is not None:
	r=filter(strip,split(src,'\0'))
	if not r: raise ValueError, 'null sql'
	if len(r) == 1:
	  r = [src]
	res=None
	try:
	  for s in r:
	    oRS = self.executeQuery(s)
	    if oRS is not None:
	      if res is not None:
		# close ResultSet object and release it
		oRS.Close()
		oRS = None
		raise ValueError, (
		    'multiple selects are not allowed')
	      res = self.prepareResultSet(oRS, rdb)
	      # close ResultSet object and release it
	      oRS.Close()
	      oRS = None
	finally:
	  pass

    except error, v:
      state, native, mess = v
      raise error, "%s (%s)" % (mess, state)

    if res==None:
      if rdb: return u'x\n8s\n'.encode('iso8859-1')
      else: return ()
      
    if rdb:
      res = join(res,u'\n')
      res += u'\n'

    return res.encode('iso8859-1')

  def executeQuery(self, sqlString):
    """Executes a sql query and returns the resultSet"""


    # Excecute query
    (oRS, rowsAffected) = self.connection.Execute(sqlString)
    if oRS.EOF and oRS.BOF: # empty RecordSet
      # close ResultSet object and release it
      oRS.Close()
      oRS = None
      return None
    else:
      return oRS
    
  def prepareResultSet(self, oRS, rdb):
    """Takes an ADO ResultSet Object and prepares it for Zope"""
    
    ncol = oRS.Fields.Count
    r=[]
    row=[u'']*ncol
    names=[u'']*ncol
    rdbdefs=[u'']*ncol
    indexes=range(ncol)
    binary_flags=[0]*ncol
    for i in range(ncol):
      name = oRS.Fields(i).Name # is unicode already
      type = oRS.Fields(i).Type
      prec = oRS.Fields(i).Precision
      scale = oRS.Fields(i).NumericScale
      nullable = ((oRS.Fields(i).Attributes & adFldIsNullable) == adFldIsNullable
		  or (oRS.Fields(i).Attributes & adFldMayBeNull) == adFldMayBeNull)
      names[i]=name
      if prec < 1: prec=8
      binary_flags[i]=type in binary_types
      if defs.has_key(type): type=defs[type]
      else: type = u't'
      if type == u'n' and scale == 0: type = u'i'
      rdbdefs[i] = type
      row[i] = u'%s%s' % (prec,type)

    if rdb:
      r.append(join(names,u'\t'))
      r.append(join(row,u'\t'))

    while not oRS.EOF:
      for i in range(ncol):
	v=oRS.Fields(i).Value
	if rdbdefs[i] == u't' and v is not None:
	  try:
	    v = unicode(v)
	  except:
	    v = unicode(v, 'iso8859-1')
	  if find(v,u'\\') >= 0: v=join(split(v,u'\\'),u'\\\\')
	  if find(v,u'\n') >= 0: v=join(split(v,u'\n'),u'\\n')
	  if find(v,u'\t') >= 0: v=join(split(v,u'\t'),u'\\t')
	elif rdbdefs[i] == u'd' and v is not None:
	  v = unicode(v.Format(u'%Y/%m/%d %H:%M:%S'))
	elif v == None:
	  v = u''
	else:
	  if rdbdefs[i] == u'n':
	    v = cvtFloat(v)
	  v = str(v)
	  v = unicode(v)
	row[i]=v
      if rdb: rd=join(row,u'\t')
      else:
	rd={}
	for i in indexes: rd[names[i]]=row[i]
      r.append(rd)
      oRS.MoveNext()

    return r

  def _begin(self):
    """Begins a transaction"""
    pass

  def _finish(self):
    """Commits a transaction"""
    # close database connection
    self.disconnect()
    
  def _abort(self):
    """Aborts a transaction"""
    # close database connection
    self.disconnect()


def cvtFloat(variant):
  """Convert a text string to a float value.
  Special care is taken of . and , as decimal separators"""
  try:
    return float(variant)
  except (ValueError,TypeError):
    pass
  if type(variant)==types.TupleType:
    return cvtMoney(variant)
  try:
    europeVsUS=translate(str(variant),maketrans(".,",",."))
    return float(europeVsUS)
  except (ValueError,TypeError):
    pass    
